# -*- coding: utf-8 -*-
"""
Build model from data contained by a folder

USAGE

    python -m thyrsis.database.build_model [-o dot_thyrsis_path] path_to_data_folder
"""

import os
import sys
import shutil
import unittest
import time
import subprocess
import getopt
from distutils.spawn import find_executable

from osr import SpatialReference
from qgis.core import (
    QgsApplication,
    QgsSettings,
    QgsVectorLayer,
    QgsFeatureRequest,
    NULL,
    QgsProject,
    QgsTransactionGroup,
    QgsFeature,
    QgsGeometry,
    QgsWkbTypes,
    QgsCoordinateReferenceSystem
)
from qgis.gui import QgisInterface, QgsMapCanvas
from qgis.utils import iface, initInterface

from qgis.PyQt.QtCore import QDate, QTime, QDateTime, QVariant
from qgis.PyQt.QtSql import QSqlDatabase, QSqlQuery

from thyrsis import database
from thyrsis.database import sqlite, load_chemicals, load_forages, update_mesh, update_mesh2, delete_mesh
from thyrsis.settings import Settings
from thyrsis.plugin import Plugin
from thyrsis.mesh_project import MeshProject
from thyrsis.gui.task_manager import Task
from thyrsis.log import logger

class DummyInterface(object):
    def __init__(self, canvas):
        self.canvas = canvas

    def __getattr__(self, *args, **kwargs):
        def dummy(*args, **kwargs):
            return self
        return dummy
    def __iter__(self):
        return self
    def next(self):
        raise StopIteration
    def layers(self):
        # simulate iface.legendInterface().layers()
        return QgsMapLayerRegistry.instance().mapLayers().values()

    def newProject(self):
        QgsProject.instance().clear()

    def mapCanvas(self):
        return self.canvas

class Model:
    def __init__(self, folder, param, contours, mnt, mur, hynverse_param, points_fixes=None, potentiel=None, points_pilote=None, permeability_raster=None, const_perm=None):
        self.site_db = os.path.join(folder, '.thyrsis', 'sites.sqlite')
        with open(param) as file :
            for line in file.readlines():
                key,value = line.split(':')
                if key=='dbname':
                    self.dbname = value.replace('\n','')
                elif key=='srid':
                    self.srid = int(value.replace('\n',''))
                elif key=='codeHydro':
                    self.codeHydro = value.replace('\n','')
        self.site = self.dbname.split('_')[0]
        self.mesh_db = os.path.join(folder, '.thyrsis', 'mesh', self.dbname+'.mesh.sqlite')
        if os.path.exists(self.mesh_db):
            os.remove(self.mesh_db)
        self.canvas = QgsMapCanvas()
        self.iface = DummyInterface(self.canvas)
        self.thyrsis = Plugin(self.iface)
        self.thyrsis.initGui()

        self.contours = contours
        self.mnt = mnt
        self.mur = mur
        self.hynverse_param = hynverse_param
        self.points_fixes = points_fixes
        self.potentiel = potentiel        

        if points_pilote :
            self.method = 1
            self.data = points_pilote
        elif permeability_raster :
            self.method = 2
            self.data = permeability_raster
        else :
            self.method = 3
            self.data = const_perm

    def createMeshProject(self):
        print('==Create Mesh Project==')
        with sqlite.connect(self.site_db, "ThyrsisBuildModel") as conn:
            cur = conn.cursor()
            existing_sites = cur.execute("SELECT id,nom from sites").fetchall()
            ids = [site[0] for site in existing_sites]
            names = [site[1] for site in existing_sites]
            if not self.site in names :
                cur.execute("INSERT INTO sites(id,nom) values ({},'{}')".format(max(ids if len(ids)>0 else [-1] )+1,self.site))
                conn.commit()

        proj = SpatialReference()
        proj.ImportFromEPSG(self.srid)
        QgsProject.instance().setCrs(QgsCoordinateReferenceSystem(proj.ExportToWkt()))
        meshProject = MeshProject.create(self.mesh_db, self.canvas)

    def createDomain(self):
        print('==Create Domain==')
        meshProject = MeshProject(self.mesh_db)
        self.thyrsis.project = meshProject
        lyr_contours = QgsVectorLayer(self.contours, 'contours_in', 'ogr')
        features = [ft for ft in lyr_contours.getFeatures()]

        db_lyr_contours = QgsProject.instance().mapLayersByName('contours')[0]
        for i in features :
            ft = QgsFeature(db_lyr_contours.fields())
            ft.setGeometry(i.geometry())
            attr = []
            for fld in db_lyr_contours.fields():
                if fld.name() == 'OGC_FID':
                    attr.append(None)
                elif fld.name() in lyr_contours.fields().names():
                    attr.append(i.attribute(fld.name()))
                else :
                    attr.append(None)
            ft.setAttributes(attr)
            db_lyr_contours.startEditing()
            db_lyr_contours.addFeatures([ft])
            db_lyr_contours.commitChanges()

        if self.points_fixes:
            lyr_points_fixe = QgsVectorLayer('file:///'+self.points_fixes+'?delimiter=%s&crs=epsg:%s&xField=%s&yField=%s' % (",", str(self.srid), "X", "Y"),'points_fixe','delimitedtext')
            features = [ft for ft in lyr_points_fixe.getFeatures()]
            db_lyr_pf = QgsProject.instance().mapLayersByName('points_fixes')[0]
            for i in features :
                ft = QgsFeature(db_lyr_pf.fields())
                ft.setGeometry(i.geometry())
                attr = []
                for fld in db_lyr_pf.fields():
                    if fld.name() == 'OGC_FID':
                        attr.append(None)
                    elif fld.name() in lyr_points_fixe.fields().names():
                        attr.append(i.attribute(fld.name()))
                    else :
                        attr.append(None)
                ft.setAttributes(attr)
                db_lyr_pf.startEditing()
                db_lyr_pf.addFeatures([ft])
                db_lyr_pf.commitChanges()

    def createMesh(self):
        print('==Create Mesh==')
        meshProject = MeshProject(self.mesh_db)
        self.thyrsis.project = meshProject
        meshProject.mesh(self.mnt, self.mur, self.potentiel)

    def createModel(self):
        print('==Create Model==')
        with open(self.hynverse_param) as file, sqlite.connect(self.mesh_db) as conn :
            cur = conn
            for line in file:
                key = line.split(':')[0]
                value = line.split(':')[1]
                cur.execute("UPDATE hynverse_parametres set {}={}".format(key,value))
            conn.commit()

        self.settings = Settings(os.path.dirname(self.mesh_db))
        self.settings.setValue('General', 'codeHydro', self.codeHydro)
        codeHydroPath = find_executable(self.codeHydro)
        self.settings.setValue('General', 'codeHydroCommand', codeHydroPath)
        self.settings.save()

        if self.method == 1 :

            lyr_point_pilot = QgsVectorLayer('file:///'+self.data+'?delimiter=%s&crs=epsg:%s&xField=%s&yField=%s' % (",", str(self.srid), "X", "Y"),'points_pilote','delimitedtext')
            features = [ft for ft in lyr_point_pilot.getFeatures()]
            db_lyr_pp = QgsProject.instance().mapLayersByName('points_pilote')[0]
            
            for i in features :
                ft = QgsFeature(db_lyr_pp.fields())
                ft.setGeometry(i.geometry())
                attr = []
                for fld in db_lyr_pp.fields():
                    if fld.name() == 'OGC_FID':
                        attr.append(None)
                    elif fld.name() in lyr_point_pilot.fields().names():
                        attr.append(i.attribute(fld.name()))
                    else :
                        attr.append(None)
                ft.setAttributes(attr)
                db_lyr_pp.startEditing()
                db_lyr_pp.addFeatures([ft])
                db_lyr_pp.commitChanges()
            
            self.thyrsis.inverse(test=True)

        elif self.method == 2:
            self.thyrsis.model_no_inverse(param=self.data, const=False, test=True)
        else :
            self.thyrsis.model_no_inverse(param=self.data, const=True, test=True)

        QgsProject.instance().write()

    def buildModel(self):
        self.createMeshProject()
        self.createDomain()
        self.createMesh()
        self.createModel()

if __name__ == "__main__":

    logger.enable_console(True)

    try:
        optlist, args = getopt.getopt(sys.argv[1:],
                "o:h",
                ["help"])
    except Exception as e:
        sys.stderr.write(str(e)+"\n")
        exit(1)
    optlist=dict(optlist)

    if len(sys.argv) == 2 and sys.argv[1] in ["-h","--help"]:
        help(sys.modules[__name__])
        exit(0)

    if len(sys.argv) < 2:
        print("wrong number of arguments (try database.py --help)")
        exit(1)

    if os.path.isdir(sys.argv[-1]):
        folderpath = sys.argv[-1]
    else :
        print("The folder doesn't exist")
        exit(1)
    param = os.path.join(os.path.abspath(folderpath),'param.txt').replace('\n','')
    print(param)
    if not os.path.isfile(param):
        print("no parameter file found in the folder")
        exit(1)
    contours = os.path.abspath(os.path.join(folderpath,'contours.gpkg'))
    if not os.path.exists(contours):
        print("No contours.gpkg file found in the folder")
        exit(1)
    mnt = os.path.abspath(os.path.join(folderpath,'mnt.tif'))
    if not os.path.exists(mnt):
        print("No mnt.tif' file found in the folder")
        exit(1)
    mur = os.path.abspath(os.path.join(folderpath,'mur.tif'))
    if not os.path.exists(mur):
        print("No mur.tif' file found in the folder")
        exit(1)
    potentiel = os.path.abspath(os.path.join(folderpath, 'potentiel.tif'))
    if not os.path.exists(potentiel):
        potentiel = None
    hynverse_param = os.path.abspath(os.path.join(folderpath,'hynverse_param.txt'))
    if not os.path.exists(hynverse_param):
        print("no parameter file found in the folder")
        exit(1)
    points_pilote = os.path.abspath(os.path.join(folderpath,'points_pilote.csv'))
    if not os.path.exists(points_pilote):
        points_pilote = None
    points_fixe = os.path.abspath(os.path.join(folderpath,'points_fixe.csv'))
    if not os.path.exists(points_fixe):
        points_fixe = None
    permeability_raster = os.path.abspath(os.path.join(folderpath,'permeabilite.tif'))
    if not os.path.exists(permeability_raster):
        permeability_raster = None
    const_perm = os.path.abspath(os.path.join(folderpath,'const_perm.txt'))
    if not os.path.exists(const_perm):
        const_perm = None
    else :
        with open(const_perm, 'r') as file :
            const_perm = float(file.readline())

    if "-o" in optlist :
        dot_thyrsis_path= str(optlist["-o"])
    else :
        dot_thyrsis_path = folderpath
    
    if (const_perm is None and permeability_raster is None and points_pilote is None):
        print("No information to build model (points_pilote.csv / permeabilite.tif / const_perm.txt)")
        exit(1)
    elif (bool(const_perm)+bool(permeability_raster)+bool(points_pilote) !=1):
        print("More than one information to build model (points_pilote.csv / permeabilite.tif / const_perm.txt). Can't decide how build the model")
        exit(1)
    else :
        if os.path.exists(os.path.join(folderpath,'.local')):
            shutil.rmtree(os.path.join(folderpath,'.local'))
        if os.path.exists(os.path.join(folderpath,'.thyrsis')):
            shutil.rmtree(os.path.join(folderpath,'.thyrsis'))
        shutil.copytree(os.path.join(os.environ['HOME'],'.local','share','QGIS','QGIS3','profiles','default'),os.path.join(folderpath,'.local','share','QGIS','QGIS3','profiles','default'))
        os.environ['HOME']=folderpath

        qgis_path = find_executable('qgis')
        if qgis_path == None :
            sys.exit("No qgis executable found, please add qgis folder in your PATH")

        os.environ['QGIS_PREFIX_PATH'] = os.path.join(os.path.dirname(qgis_path),'..')
        ld_lib_path = os.pathsep + os.environ['LD_LIBRARY_PATH'] if 'LD_LIBRARY_PATH' in os.environ else ''
        os.environ['LD_LIBRARY_PATH'] = os.path.join(os.path.dirname(qgis_path),'..','lib')+ ld_lib_path
        os.environ['QGIS_PLUGIN_PATH'] = os.path.join(os.path.dirname(qgis_path),'..','python','plugins')
        # QgsApplication.setPrefixPath(os.path.join(os.path.dirname(qgis_path),'..'))

        qgisApp = QgsApplication([], True)
        #qgisApp.dataItemProviderRegistry().addProvider(QgsSpatiaLiteProvider)
        qgisApp.initQgis()

        model = Model(dot_thyrsis_path, param, contours, mnt, mur, hynverse_param, points_fixe, potentiel, points_pilote, permeability_raster, const_perm)
        model.buildModel()